#include "int_da.h"
#include <stdio.h>
#include <stdlib.h>

// Defining this enables an optimization for when the capacity is 0 that
// leaves the data pointer NULL rather than heap allocating 0 bytes.
#define ZERO_CAP_OPTIMIZATION

typedef int element_t;

struct int_da
{
    size_t capacity;
    size_t size;
    element_t* data;
};

int_da_t int_da_create(size_t capacity)
{
    int_da_t result = malloc(sizeof(struct int_da));
    if (result == NULL) goto fail1;

    result->capacity = capacity;
    result->size = 0;

#ifdef ZERO_CAP_OPTIMIZATION
    if (capacity == 0) {
        result->data = NULL;
        return result;
    }
#endif

    result->data = malloc(capacity * sizeof(element_t));
    if (result->data == NULL) goto fail2;

    return result;

    // These labels are a common pattern for C error handling; it's sometimes
    // known as a "goto chain." When we perform a sequence of allocations,
    // any of which might fail, then we need to be able to undo them in reverse
    // order from any point. (Usually a chain with only two labels isn't
    // worth it.)

    fail2:
    free(result);

    fail1:
    return NULL;
}

void int_da_destroy(int_da_t da)
{
    free(da->data);
    free(da);
}

size_t int_da_size(int_da_t da)
{
    return da->size;
}

element_t* int_da_index(int_da_t da, size_t ix)
{
    return &da->data[ix];
}

static size_t max(size_t a, size_t b)
{
    return a < b ? b : a;
}

element_t* int_da_push_back(int_da_t da)
{
    if (da->size == da->capacity) {
        size_t new_cap = max(da->size + 1, 2 * da->capacity);
        element_t* new_data = realloc(da->data, new_cap * sizeof(element_t));
        if (new_data == NULL) return NULL;

        da->capacity = new_cap;
        da->data = new_data;
    }

    return &da->data[da->size++];
}

void int_da_pop_back(int_da_t da)
{
    --da->size;
}

